"""GS1 messages."""

from __future__ import annotations

import re
from dataclasses import dataclass
from itertools import chain
from typing import Iterable, List, Optional, Union

from biip import ParseError
from biip.gs1 import (
    ASCII_GROUP_SEPARATOR,
    DEFAULT_SEPARATOR_CHARS,
    GS1ApplicationIdentifier,
    GS1ElementString,
)
from biip.gs1._application_identifiers import _GS1_APPLICATION_IDENTIFIERS
from biip.gtin import RcnRegion


@dataclass
class GS1Message:
    """A GS1 message is the result of a single barcode scan.

    It may contain one or more GS1 Element Strings.

    Example:
        See :mod:`biip.gs1` for a usage example.
    """

    #: Raw unprocessed value.
    value: str

    #: List of Element Strings found in the message.
    element_strings: List[GS1ElementString]

    @classmethod
    def parse(
        cls,
        value: str,
        *,
        rcn_region: Optional[RcnRegion] = None,
        rcn_verify_variable_measure: bool = True,
        separator_chars: Iterable[str] = DEFAULT_SEPARATOR_CHARS,
    ) -> GS1Message:
        """Parse a string from a barcode scan as a GS1 message with AIs.

        Args:
            value: The string to parse.
            rcn_region: The geographical region whose rules should be used to
                interpret Restricted Circulation Numbers (RCN).
                Needed to extract e.g. variable weight/price from GTIN.
            rcn_verify_variable_measure: Whether to verify that the variable
                measure in a RCN matches its check digit, if present. Some
                companies use the variable measure check digit for other
                purposes, requiring this check to be disabled.
            separator_chars: Characters used in place of the FNC1 symbol.
                Defaults to `<GS>` (ASCII value 29).
                If variable-length fields in the middle of the message are
                not terminated with a separator character, the parser might
                greedily consume the rest of the message.

        Returns:
            A message object with one or more element strings.

        Raises:
            ParseError: If a fixed-length field ends with a separator character.
        """
        value = value.strip()
        element_strings = []
        rest = value[:]

        while rest:
            element_string = GS1ElementString.extract(
                rest,
                rcn_region=rcn_region,
                rcn_verify_variable_measure=rcn_verify_variable_measure,
                separator_chars=separator_chars,
            )
            element_strings.append(element_string)

            rest = rest[len(element_string) :]

            if rest.startswith(tuple(separator_chars)):
                if element_string.ai.fnc1_required:
                    rest = rest[1:]
                else:
                    separator_char = rest[0]
                    raise ParseError(
                        f"Element String {element_string.as_hri()!r} has fixed length "
                        "and should not end with a separator character. "
                        f"Separator character {separator_char!r} found in {value!r}."
                    )

        return cls(value=value, element_strings=element_strings)

    @classmethod
    def parse_hri(
        cls,
        value: str,
        *,
        rcn_region: Optional[RcnRegion] = None,
        rcn_verify_variable_measure: bool = True,
    ) -> GS1Message:
        """Parse the GS1 string given in HRI (human readable interpretation) format.

        Args:
            value: The HRI string to parse.
            rcn_region: The geographical region whose rules should be used to
                interpret Restricted Circulation Numbers (RCN).
                Needed to extract e.g. variable weight/price from GTIN.
            rcn_verify_variable_measure: Whether to verify that the variable
                measure in a RCN matches its check digit, if present. Some
                companies use the variable measure check digit for other
                purposes, requiring this check to be disabled.

        Returns:
            A message object with one or more element strings.

        Raises:
            ParseError: If parsing of the data fails.
        """
        value = value.strip()
        if not value.startswith("("):
            raise ParseError(
                f"Expected HRI string {value!r} to start with a parenthesis."
            )

        pattern = r"\((\d+)\)(\w+)"
        matches = re.findall(pattern, value)
        if not matches:
            raise ParseError(
                f"Could not find any GS1 Application Identifiers in {value!r}. "
                "Expected format: '(AI)DATA(AI)DATA'."
            )

        pairs = []
        for ai_number, ai_data in matches:
            if ai_number not in _GS1_APPLICATION_IDENTIFIERS:
                raise ParseError(
                    f"Unknown GS1 Application Identifier {ai_number!r} in {value!r}."
                )
            pairs.append((_GS1_APPLICATION_IDENTIFIERS[ai_number], ai_data))

        parts = chain(
            *[
                [
                    gs1_ai.ai,
                    ai_data,
                    (ASCII_GROUP_SEPARATOR if gs1_ai.fnc1_required else ""),
                ]
                for gs1_ai, ai_data in pairs
            ]
        )
        normalized_string = "".join(parts)
        return GS1Message.parse(
            normalized_string,
            rcn_region=rcn_region,
            rcn_verify_variable_measure=rcn_verify_variable_measure,
        )

    def as_hri(self) -> str:
        """Render as a human readable interpretation (HRI).

        The HRI is often printed directly below barcodes.

        Returns:
            A human-readable string where the AIs are wrapped in parenthesis.
        """
        return "".join(es.as_hri() for es in self.element_strings)

    def filter(
        self,
        *,
        ai: Optional[Union[str, GS1ApplicationIdentifier]] = None,
        data_title: Optional[str] = None,
    ) -> List[GS1ElementString]:
        """Filter Element Strings by AI or data title.

        Args:
            ai: AI instance or string to match against the start of the
                Element String's AI.
            data_title: String to find anywhere in the Element String's AI
                data title.

        Returns:
            All matching Element Strings in the message.
        """
        if isinstance(ai, GS1ApplicationIdentifier):
            ai = ai.ai

        result = []

        for element_string in self.element_strings:
            if ai is not None and element_string.ai.ai.startswith(ai):
                result.append(element_string)
            elif data_title is not None and data_title in element_string.ai.data_title:
                result.append(element_string)

        return result

    def get(
        self,
        *,
        ai: Optional[Union[str, GS1ApplicationIdentifier]] = None,
        data_title: Optional[str] = None,
    ) -> Optional[GS1ElementString]:
        """Get Element String by AI or data title.

        Args:
            ai: AI instance or string to match against the start of the
                Element String's AI.
            data_title: String to find anywhere in the Element String's AI
                data title.

        Returns:
            The first matching Element String in the message.
        """
        matches = self.filter(ai=ai, data_title=data_title)
        return matches[0] if matches else None



"""Support for barcode data with GS1 element strings.

The :mod:`biip.gs1` module contains Biip's support for parsing data
consisting of GS1 Element Strings. Each Element String is identified by a GS1
Application Identifier (AI) prefix.

Data of this format is found in the following types of barcodes:

- GS1-128
- GS1 DataBar
- GS1 DataMatrix
- GS1 QR Code

If you only want to parse GS1 Messages, you can import the GS1 Message parser
directly instead of using :func:`biip.parse`.

    >>> from biip.gs1 import GS1Message

If the parsing succeeds, it returns a :class:`GS1Message` object.

    >>> msg = GS1Message.parse("010703206980498815210526100329")

The ``GS1Message`` has a raw value as well as an HRI, short for "human readable
interpretation". The HRI is the text usually printed below or next to the
barcode.

    >>> msg.value
    '010703206980498815210526100329'
    >>> msg.as_hri()
    '(01)07032069804988(15)210526(10)0329'

HRI can also be parsed.

    >>> GS1Message.parse_hri("(01)07032069804988(15)210526(10)0329")

A message can contain multiple element strings.

    >>> len(msg.element_strings)
    3

In this example, the first element string is a GTIN.

    >>> msg.element_strings[0]
    GS1ElementString(ai=GS1ApplicationIdentifier(ai='01', description='Global
    Trade Item Number (GTIN)', data_title='GTIN', fnc1_required=False,
    format='N2+N14'), value='07032069804988', pattern_groups=['07032069804988'],
    gln=None, gln_error=None, gtin=Gtin(value='07032069804988',
    format=GtinFormat.GTIN_13, prefix=GS1Prefix(value='703', usage='GS1
    Norway'), company_prefix=GS1CompanyPrefix(value='703206'),
    payload='703206980498', check_digit=8, packaging_level=None),
    gtin_error=None, sscc=None, sscc_error=None, date=None, decimal=None,
    money=None)

The message object has :meth:`~GS1Message.get` and :meth:`~GS1Message.filter`
methods to lookup element strings either by the Application Identifier's
"data title" or its AI number.

    >>> msg.get(data_title='BEST BY')
    GS1ElementString(ai=GS1ApplicationIdentifier(ai='15', description='Best
    before date (YYMMDD)', data_title='BEST BEFORE or BEST BY',
    fnc1_required=False, format='N2+N6'), value='210526',
    pattern_groups=['210526'], gln=None, gln_error=None, gtin=None,
    gtin_error=None, sscc=None, sscc_error=None, date=datetime.date(2021, 5,
    26), decimal=None, money=None)
    >>> msg.get(ai="10")
    GS1ElementString(ai=GS1ApplicationIdentifier(ai='10', description='Batch
    or lot number', data_title='BATCH/LOT', fnc1_required=True,
    format='N2+X..20'), value='0329', pattern_groups=['0329'], gln=None,
    gln_error=None, gtin=None, gtin_error=None, sscc=None, sscc_error=None,
    date=None, decimal=None, money=None)
"""

from typing import Tuple

ASCII_GROUP_SEPARATOR = "\x1d"

#: The default separator character is <GS>, ASCII value 29.
#:
#: References:
#:   GS1 General Specifications, section 7.8.3.
DEFAULT_SEPARATOR_CHARS: Tuple[str] = (ASCII_GROUP_SEPARATOR,)

# The following must be imported in this specific order.
from biip.gs1._symbology import GS1Symbology  # isort:skip  # noqa: E402
from biip.gs1._application_identifiers import (  # isort:skip  # noqa: E402
    GS1ApplicationIdentifier,
)
from biip.gs1._prefixes import GS1CompanyPrefix, GS1Prefix  # isort:skip  # noqa: E402
from biip.gs1._element_strings import (  # isort:skip  # noqa: E402
    GS1ElementString,
)
from biip.gs1._messages import GS1Message  # isort:skip  # noqa: E402

__all__ = [
    "GS1CompanyPrefix",
    "GS1Message",
    "GS1ElementString",
    "GS1ApplicationIdentifier",
    "GS1Prefix",
    "GS1Symbology",
    "DEFAULT_SEPARATOR_CHARS",
]



r"""Biip interprets the data in barcodes.

    >>> import biip

An ambiguous value may be interpreted as different formats. In the following
example, the value can be interpreted as either a GTIN or a GS1 Message.

    >>> result = biip.parse("96385074")
    >>> result.gtin
    Gtin(value='96385074', format=GtinFormat.GTIN_8,
    prefix=GS1Prefix(value='00009', usage='GS1 US'),
    company_prefix=GS1CompanyPrefix(value='0000963'), payload='9638507',
    check_digit=4, packaging_level=None)
    >>> result.gs1_message
    GS1Message(value='96385074',
    element_strings=[GS1ElementString(ai=GS1ApplicationIdentifier(ai='96',
    description='Company internal information', data_title='INTERNAL',
    fnc1_required=True, format='N2+X..90'), value='385074',
    pattern_groups=['385074'], gln=None, gln_error=None, gtin=None,
    gtin_error=None, sscc=None, sscc_error=None, date=None, decimal=None,
    money=None)])

In the next example, the value is only valid as a GS1 Message and the GTIN
parser returns an error explaining why the value cannot be interpreted as a
GTIN. If a format includes check digits, Biip always control them and fail if
the check digits are incorrect.

    >>> result = biip.parse("15210527")
    >>> result.gtin
    None
    >>> result.gtin_error
    "Invalid GTIN check digit for '15210527': Expected 4, got 7."
    >>> result.gs1_message
    GS1Message(value='15210527',
    element_strings=[GS1ElementString(ai=GS1ApplicationIdentifier(ai='15',
    description='Best before date (YYMMDD)', data_title='BEST BEFORE or BEST
    BY', fnc1_required=False, format='N2+N6'), value='210527',
    pattern_groups=['210527'], gln=None, gln_error=None, gtin=None,
    gtin_error=None, sscc=None, sscc_error=None, date=datetime.date(2021, 5,
    27), decimal=None, money=None)])

If a value cannot be interpreted as any supported format, an exception is
raised with a reason from each parser.

    >>> biip.parse("123")
    Traceback (most recent call last):
        ...
    biip._exceptions.ParseError: Failed to parse '123':
    - GTIN: Failed to parse '123' as GTIN: Expected 8, 12, 13, or 14 digits, got 3.
    - UPC: Failed to parse '123' as UPC: Expected 6, 7, 8, or 12 digits, got 3.
    - SSCC: Failed to parse '123' as SSCC: Expected 18 digits, got 3.
    - GS1: Failed to match '123' with GS1 AI (12) pattern '^12(\d{6})$'.
"""

try:
    from importlib.metadata import PackageNotFoundError, version  # type: ignore
except ImportError:  # pragma: no cover
    from importlib_metadata import version, PackageNotFoundError  # type: ignore

from biip._exceptions import BiipException, EncodeError, ParseError
from biip._parser import ParseResult, parse

__all__ = [
    "parse",
    "ParseResult",
    "BiipException",
    "EncodeError",
    "ParseError",
]


try:
    __version__ = version(__name__)
except PackageNotFoundError:  # pragma: no cover
    __version__ = "unknown"



"""Serial Shipping Container Code (SSCC).

SSCCs are used to identify logistic units, e.g. a pallet shipped between two
parties.

If you only want to parse SSCCs, you can import the SSCC parser directly
instead of using :func:`biip.parse`.

    >>> from biip.sscc import Sscc

If parsing succeeds, it returns a :class:`Sscc` object.

    >>> sscc = Sscc.parse("157035381410375177")
    >>> sscc
    Sscc(value='157035381410375177', prefix=GS1Prefix(value='570', usage='GS1
    Denmark'), company_prefix=GS1CompanyPrefix(value='5703538'),
    extension_digit=1, payload='15703538141037517', check_digit=7)

Biip can format the SSCC in HRI format for printing on a label.

    >>> sscc.as_hri()
    '1 5703538 141037517 7'

If the detected GS1 Company Prefix length is wrong, it can be overridden:

    >>> sscc.as_hri(company_prefix_length=9)
    '1 570353814 1037517 7'
"""

from __future__ import annotations

from dataclasses import dataclass
from typing import Optional

from biip import ParseError
from biip.gs1 import GS1CompanyPrefix, GS1Prefix
from biip.gs1.checksums import numeric_check_digit


@dataclass
class Sscc:
    """Data class containing an SSCC."""

    #: Raw unprocessed value.
    value: str

    #: The GS1 Prefix, indicating what GS1 country organization that assigned
    #: code range.
    prefix: Optional[GS1Prefix]

    #: The GS1 Company Prefix, identifying the company that issued the SSCC.
    company_prefix: Optional[GS1CompanyPrefix]

    #: Extension digit used to increase the capacity of the serial reference.
    extension_digit: int

    #: The actual payload, including extension digit, company prefix, and item
    #: reference. Excludes the check digit.
    payload: str

    #: Check digit used to check if the SSCC as a whole is valid.
    check_digit: int

    @classmethod
    def parse(cls, value: str) -> Sscc:
        """Parse the given value into a :class:`Sscc` object.

        Args:
            value: The value to parse.

        Returns:
            SSCC data structure with the successfully extracted data.
            The checksum is guaranteed to be valid if an SSCC object is returned.

        Raises:
            ParseError: If the parsing fails.
        """
        value = value.strip()

        if len(value) != 18:
            raise ParseError(
                f"Failed to parse {value!r} as SSCC: "
                f"Expected 18 digits, got {len(value)}."
            )

        if not value.isdecimal():
            raise ParseError(
                f"Failed to parse {value!r} as SSCC: Expected a numerical value."
            )

        value_without_extension_digit = value[1:]
        prefix = GS1Prefix.extract(value_without_extension_digit)
        company_prefix = GS1CompanyPrefix.extract(value_without_extension_digit)
        extension_digit = int(value[0])
        payload = value[:-1]
        check_digit = int(value[-1])

        calculated_check_digit = numeric_check_digit(payload)
        if check_digit != calculated_check_digit:
            raise ParseError(
                f"Invalid SSCC check digit for {value!r}: "
                f"Expected {calculated_check_digit!r}, got {check_digit!r}."
            )

        return cls(
            value=value,
            prefix=prefix,
            company_prefix=company_prefix,
            extension_digit=extension_digit,
            payload=payload,
            check_digit=check_digit,
        )

    def as_hri(self, *, company_prefix_length: Optional[int] = None) -> str:
        """Render as a human readable interpretation (HRI).

        The HRI is often printed directly below barcodes.

        The GS1 Company Prefix length will be detected and used to render the
        Company Prefix and the Serial Reference as two separate groups. If the
        GS1 Company Prefix length cannot be found, the Company Prefix and the
        Serial Reference are rendered as a single group.

        Args:
            company_prefix_length: Override the detected GS1 Company Prefix
                length. 7-10 characters. If not specified, the GS1 Company
                Prefix is automatically detected.

        Raises:
            ValueError: If an illegal company prefix length is given.

        Returns:
            A human-readable string where the logic parts are separated by whitespace.
        """
        value = self.payload[1:]  # Strip extension digit

        if company_prefix_length is not None:
            # Using override of GS1 Company Prefix length
            if not (7 <= company_prefix_length <= 10):
                raise ValueError(
                    "Expected company prefix length between 7 and 10, "
                    f"got {company_prefix_length!r}."
                )
        elif self.company_prefix is not None:
            # Using auto-detected GS1 Company Prefix length
            company_prefix_length = len(self.company_prefix.value)

        if company_prefix_length is None:
            return f"{self.extension_digit} {value} {self.check_digit}"

        company_prefix = value[:company_prefix_length]
        serial = value[company_prefix_length:]
        return f"{self.extension_digit} {company_prefix} {serial} {self.check_digit}"



"""Support for Global Trade Item Number (GTIN).

The :mod:`biip.gtin` module contains Biip's support for parsing GTIN formats.

A GTIN is a number that uniquely identifies a trade item.

This class can interpet the following GTIN formats:

- GTIN-8, found in EAN-8 barcodes.
- GTIN-12, found in UPC-A and UPC-E barcodes.
- GTIN-13, found in EAN-13 barcodes.
- GTIN-14, found in ITF-14 barcodes, as well as a data field in GS1 barcodes.

If you only want to parse GTINs, you can import the GTIN parser directly
instead of using :func:`biip.parse`.

    >>> from biip.gtin import Gtin

If parsing succeeds, it returns a :class:`Gtin` object.

    >>> gtin = Gtin.parse("7032069804988")
    >>> gtin
    Gtin(value='7032069804988', format=GtinFormat.GTIN_13,
    prefix=GS1Prefix(value='703', usage='GS1 Norway'),
    company_prefix=GS1CompanyPrefix(value='703206'), payload='703206980498',
    check_digit=8, packaging_level=None)

A GTIN can be converted to any other GTIN format, as long as the target
format is longer.

    >>> gtin.as_gtin_14()
    '07032069804988'

As all GTINs can be converted to GTIN-14, it is the recommended format to use
when storing or comparing GTINs. For example, when looking up a product
associated with a GTIN, the GTIN should first be expanded to a GTIN-14 before
using it to query the product catalog.
"""

from biip.gtin._enums import GtinFormat, RcnRegion, RcnUsage
from biip.gtin._gtin import Gtin
from biip.gtin._rcn import Rcn

__all__ = ["Gtin", "GtinFormat", "Rcn", "RcnUsage", "RcnRegion"]



"""Enums used when parsing GTINs."""

from __future__ import annotations

from enum import Enum, IntEnum
from typing import Optional


class GtinFormat(IntEnum):
    """Enum of GTIN formats."""

    #: GTIN-8
    GTIN_8 = 8

    #: GTIN-12
    GTIN_12 = 12

    #: GTIN-13
    GTIN_13 = 13

    #: GTIN-14
    GTIN_14 = 14

    def __str__(self) -> str:
        """Pretty string representation of format."""
        # XXX str() is a workaround of a mypy bug until
        # https://github.com/python/mypy/pull/14133 is released:
        return str(self.name).replace("_", "-")

    def __repr__(self) -> str:
        """Canonical string representation of format."""
        return f"GtinFormat.{self.name}"

    @property
    def length(self) -> int:
        """Length of a GTIN of the given format."""
        return int(self)


class RcnUsage(Enum):
    """Enum of RCN usage restrictions."""

    #: Usage of RCN restricted to geopgraphical area.
    GEOGRAPHICAL = "geo"

    #: Usage of RCN restricted to internally in a company.
    COMPANY = "company"

    def __repr__(self) -> str:
        """Canonical string representation of format."""
        return f"RcnUsage.{self.name}"


class RcnRegion(Enum):
    """Enum of geographical regions with custom RCN rules.

    The value of the enum is the lowercase ISO 3166-1 Alpha-2 code.
    """

    #: Denmark
    DENMARK = "dk"

    #: Estonia
    ESTONIA = "ee"

    #: Finland
    FINLAND = "fi"

    #: Germany
    GERMANY = "de"

    #: Great Britain
    GREAT_BRITAIN = "gb"

    #: Latvia
    LATVIA = "lv"

    #: Lithuania
    LITHUANIA = "lt"

    #: Norway
    NORWAY = "no"

    #: Sweden
    SWEDEN = "se"

    def __repr__(self) -> str:
        """Canonical string representation of format."""
        return f"RcnRegion.{self.name}"

    def get_currency_code(self) -> Optional[str]:
        """Get the ISO-4217 currency code for the region."""
        return {
            RcnRegion.DENMARK: "DKK",
            RcnRegion.GERMANY: "EUR",
            RcnRegion.GREAT_BRITAIN: "GBP",
            RcnRegion.NORWAY: "NOK",
            RcnRegion.SWEDEN: "SEK",
        }.get(self)



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['biip', 'biip.gs1', 'biip.gtin']

package_data = \
{'': ['*']}

extras_require = \
{':python_version < "3.8"': ['importlib_metadata>=1.6,<5.0'],
 'money': ['py-moneyed>=0.8']}

setup_kwargs = {
    'name': 'biip',
    'version': '3.0.0',
    'description': 'Biip interprets the data in barcodes.',
    'long_description': '# &#x1F4E6; Biip\n\n_Biip interprets the data in barcodes._\n\n[![Tests](https://img.shields.io/github/actions/workflow/status/jodal/biip/tests.yml?branch=main)](https://github.com/jodal/biip/actions/workflows/tests.yml)\n[![Docs](https://img.shields.io/readthedocs/biip)](https://biip.readthedocs.io/en/latest/)\n[![Coverage](https://img.shields.io/codecov/c/gh/jodal/biip)](https://codecov.io/gh/jodal/biip)\n[![PyPI](https://img.shields.io/pypi/v/biip)](https://pypi.org/project/biip/)\n\n---\n\nBiip is a Python library for making sense of the data in barcodes.\n\nThe library can interpret the following formats:\n\n- GTIN-8, GTIN-12, GTIN-13, and GTIN-14 numbers,\n  commonly found in EAN-8, EAN-13, and ITF-14 barcodes.\n\n- GS1 AI element strings,\n  commonly found in GS1-128 barcodes.\n\n- UPC-A and UPC-E numbers, as found in UPC-A and UPC-E barcodes.\n\nFor a quickstart guide and a complete API reference,\nsee the [documentation](https://biip.readthedocs.io/).\n\n## Installation\n\nBiip requires Python 3.7 or newer.\n\nBiip is available from [PyPI](https://pypi.org/project/biip/):\n\n```\npython3 -m pip install biip\n```\n\nOptionally, with the help of `py-moneyed`, Biip can convert amounts with\ncurrency information to `moneyed.Money` objects.\nTo install Biip with `py-moneyed`, run:\n\n```\npython3 -m pip install "biip[money]"\n```\n\n## Project resources\n\n- [Documentation](https://biip.readthedocs.io/)\n- [Source code](https://github.com/jodal/biip)\n- [Releases](https://github.com/jodal/biip/releases)\n- [Issue tracker](https://github.com/jodal/biip/issues)\n- [Contributors](https://github.com/jodal/biip/graphs/contributors)\n- [Users](https://github.com/jodal/biip/wiki/Users)\n\n## Development status\n\nAll planned features have been implemented.\nPlease open an issue if you have any barcode parsing related needs that are not covered.\n\n## Features\n\n- GS1 (multiple Element Strings with Application Identifiers)\n  - Recognize all specified Application Identifiers.\n  - Recognize allocating GS1 Member Organization from the GS1 Company Prefix.\n  - Recognize the GS1 Company Prefix.\n  - Parse fixed-length Element Strings.\n  - Parse variable-length Element Strings.\n    - Support configuring the separation character.\n  - Parse AI `00` as SSCC.\n  - Parse AI `01` and `02` as GTIN.\n  - Parse AI `410`-`417` as GLN.\n  - Parse dates into `datetime.date` values.\n    - Interpret the year to be within -49/+50 years from today.\n    - Interpret dates with day "00" as the last day of the month.\n  - Parse variable measurement fields into `Decimal` values.\n  - Parse discount percentage into `Decimal` values.\n  - Parse amounts into `Decimal` values.\n    - Additionally, if py-moneyed is installed,\n      parse amounts with currency into `Money` values.\n  - Encode as Human Readable Interpretation (HRI),\n    e.g. with parenthesis around the AI numbers.\n  - Parse Human Readable Interpretation (HRI) strings.\n  - Easy lookup of parsed Element Strings by:\n    - Application Identifier (AI) prefix\n    - Part of AI\'s data title\n- GLN (Global Location Number)\n  - Parse.\n  - Extract and validate check digit.\n  - Extract GS1 Prefix.\n  - Extract GS1 Company Prefix.\n- GTIN (Global Trade Item Number)\n  - Parse GTIN-8, e.g. from EAN-8 barcodes.\n  - Parse GTIN-12, e.g. from UPC-A and UPC-E barcodes.\n  - Parse GTIN-13, e.g. from EAN-13 barcodes.\n  - Parse GTIN-14, e.g. from ITF-14 and GS1-128 barcodes.\n  - Extract and validate check digit.\n  - Extract GS1 Prefix.\n  - Extract GS1 Company Prefix.\n  - Extract packaging level digit from GTIN-14.\n  - Encode GTIN-8 as GTIN-12/13/14.\n  - Encode GTIN-12 as GTIN-13/14.\n  - Encode GTIN-13 as GTIN-14.\n- RCN (Restricted Circulation Numbers), a subset of GTINs\n  - Classification of RCN usage to either a geographical region or a company.\n  - Parsing of variable measurements (price/weight) into `Decimal`\n    values.\n  - Parsing of price values into `Money` values if `py-moneyed` is\n    installed and the region\'s RCN parsing rules specifies a currency.\n  - Denmark: Parsing of weight and price.\n  - Estland: Parsing of weight.\n  - Finland: Parsing of weight.\n  - Germany: Parsing of weight, price, and count, including validation of\n    measurement check digit.\n  - Great Britain: Parsing of price, including validation of measurement check\n    digit.\n  - Latvia: Parsing of weight.\n  - Lithuania: Parsing of weight.\n  - Norway: Parsing of weight and price.\n  - Sweden: Parsing of weight and price.\n  - Encode RCN with the variable measure part zeroed out,\n    to help looking up the correct trade item.\n- SSCC (Serial Shipping Container Code)\n  - Extract and validate check digit.\n  - Extract GS1 Prefix.\n  - Extract GS1 Company Prefix.\n  - Extract extension digit.\n  - Encode for human consumption, with the logical groups separated by whitespace.\n- UPC (Universal Product Code)\n  - Parse 12-digit UPC-A.\n  - Parse 6-digit UPC-E, with implicit number system 0 and no check digit.\n  - Parse 7-digit UPC-E, with explicit number system and no check digit.\n  - Parse 8-digit UPC-E, with explicit number system and a check digit.\n  - Expand UPC-E to UPC-A.\n  - Suppress UPC-A to UPC-E, for the values where it is supported.\n- Symbology Identifiers, e.g. `]EO`\n  - Recognize all specified Symbology Identifier code characters.\n  - Strip Symbology Identifiers before parsing the remainder.\n  - Use Symbology Identifiers when automatically selecting what parser to use.\n\n## License\n\nBiip is copyright 2020-2022 Stein Magnus Jodal and contributors.\nBiip is licensed under the\n[Apache License, Version 2.0](https://www.apache.org/licenses/LICENSE-2.0).\n',
    'author': 'Stein Magnus Jodal',
    'author_email': 'stein.magnus@jodal.no',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/jodal/biip',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'extras_require': extras_require,
    'python_requires': '>=3.7.0,<4.0.0',
}


setup(**setup_kwargs)



"""Prefixes allocated by GS1.

Example:
    >>> from biip.gs1 import GS1CompanyPrefix, GS1Prefix
    >>> GS1Prefix.extract("7044610873466")
    GS1Prefix(value='704', usage='GS1 Norway')
    >>> GS1CompanyPrefix.extract("7044610873466")
    GS1CompanyPrefix(value='704461')
"""

from __future__ import annotations

import json
import lzma
import pathlib
from dataclasses import dataclass
from typing import Dict, Optional, Union

from biip import ParseError

_TrieNode = Union[Dict[str, "_TrieNode"], int]


@dataclass(frozen=True)
class GS1Prefix:
    """Prefix assigned by GS1.

    Used to split the allocation space of various number schemes, e.g. GTIN,
    among GS1 organizations worldwide.

    The GS1 Prefix does not identify the origin of a product, only where the
    number was assigned to a GS1 member organization.

    References:
        https://www.gs1.org/standards/id-keys/company-prefix

    Example:
        >>> from biip.gs1 import GS1Prefix
        >>> GS1Prefix.extract("978-1-492-05374-3")
        GS1Prefix(value='978', usage='Bookland (ISBN)')
    """

    #: The prefix itself.
    value: str

    #: Description of who is using the prefix.
    usage: str

    @classmethod
    def extract(cls, value: str) -> Optional[GS1Prefix]:
        """Extract the GS1 Prefix from the given value.

        Args:
            value: The string to extract a GS1 Prefix from.

        Returns:
            Metadata about the extracted prefix, or `None` if the prefix is unknown.

        Raises:
            ParseError: If the parsing fails.
        """
        for prefix_range in _GS1_PREFIX_RANGES:
            prefix = value[: prefix_range.length]

            if not prefix.isdecimal():
                continue
            number = int(prefix)

            if prefix_range.min_value <= number <= prefix_range.max_value:
                return cls(value=prefix, usage=prefix_range.usage)

        if not prefix.isdecimal():
            # `prefix` is now the shortest prefix possible, and should be
            # numeric even if the prefix assignment is unknown.
            raise ParseError(f"Failed to get GS1 Prefix from {value!r}.")

        return None


@dataclass(frozen=True)
class GS1CompanyPrefix:
    """Company prefix assigned by GS1.

    The prefix assigned to a single company.

    Example:
        >>> from biip.gs1 import GS1CompanyPrefix
        >>> GS1CompanyPrefix.extract("7044610873466")
        GS1CompanyPrefix(value='704461')
    """

    #: The company prefix itself.
    value: str

    @classmethod
    def extract(cls, value: str) -> Optional[GS1CompanyPrefix]:
        """Extract the GS1 Company Prefix from the given value.

        Args:
            value: The string to extract a GS1 Company Prefix from.
                The value is typically a GLN, GTIN, or an SSCC.

        Returns:
            Metadata about the extracted prefix, or `None` if the prefix is unknown.

        Raises:
            ParseError: If the parsing fails.
        """
        if not value.isdecimal():
            raise ParseError(f"Failed to get GS1 Company Prefix from {value!r}.")

        node = _GS1_COMPANY_PREFIX_TRIE
        digits = list(value)

        while digits:
            digit = digits.pop(0)

            assert isinstance(node, dict)
            if digit not in node:
                # Prefix is undefined.
                return None
            node = node[digit]

            if isinstance(node, int):
                if node == 0:
                    # Prefix is defined, but has zero length.
                    return None
                return cls(value=value[:node])

        return None


@dataclass(frozen=True)
class _GS1PrefixRange:
    length: int
    min_value: int
    max_value: int
    usage: str


_GS1_PREFIX_RANGES_FILE = pathlib.Path(__file__).parent / "_prefix_ranges.json"

_GS1_PREFIX_RANGES = [
    _GS1PrefixRange(**kwargs)
    for kwargs in json.loads(_GS1_PREFIX_RANGES_FILE.read_text())
]

_GS1_COMPANY_PREFIX_TRIE_FILE = (
    pathlib.Path(__file__).parent / "_company_prefix_trie.json.lzma"
)

with lzma.open(_GS1_COMPANY_PREFIX_TRIE_FILE) as fh:
    _GS1_COMPANY_PREFIX_TRIE: _TrieNode = json.load(fh)
